/*
************************************************************************
*
*   ExCalcAxis.c - CalcAxis command
*
*   Copyright (c) 1994-95
*
*   ETH Zuerich
*   Institut fuer Molekularbiologie und Biophysik
*   ETH-Hoenggerberg
*   CH-8093 Zuerich
*
*   SPECTROSPIN AG
*   Industriestr. 26
*   CH-8117 Faellanden
*
*   All Rights Reserved
*
*   Date of last modification : 96/03/17
*   Pathname of SCCS file     : /sgiext/molmol/src/cmdcalc/SCCS/s.ExCalcAxis.c
*   SCCS identification       : 1.5
*
************************************************************************
*/

#include <cmd_calc.h>

#include <stdio.h>
#include <stdlib.h>
#include <values.h>
#include <math.h>

#include <break.h>
#include <princip_axis.h>
#include <pu.h>
#include <arg.h>
#include <data_hand.h>
#include <data_sel.h>

typedef struct {
  Vec3 *coordA;
  int atomI;
} AtomData;

typedef struct {
  int molNo;
  int bondNo;
  int molI;
  int bondI;
  Vec3 vA[3];
  float *val;
  char *formRes;
} BondData;

static BOOL ListBonds = FALSE;

static PuTextWindow TextW;

static void
maxRes(DhResP resP, void *clientData)
{
  int *maxResIP = clientData;
  int resI;

  resI = DhResGetNumber(resP);
  if (resI > *maxResIP)
    *maxResIP = resI;
}

static void
countAtoms(DhAtomP atomP, void *clientData)
{
  *(int *) clientData += 1;
}

static void
fillAtom(DhAtomP atomP, void *clientData)
{
  AtomData *dataP = clientData;

  DhAtomGetCoord(atomP, dataP->coordA[dataP->atomI]);
  dataP->atomI++;
}

static void
countBonds(DhBondP atomP, DhAtomP atom1P, DhAtomP atom2P, void *clientData)
{
  *(int *) clientData += 1;
}

static void
getBondVal(DhBondP bondP, DhAtomP atom1P, DhAtomP atom2P, void *clientData)
{
  BondData *dataP = clientData;
  Vec3 x1, v;
  int axisI;
  float c;

  DhAtomGetCoord(atom1P, x1);
  DhAtomGetCoord(atom2P, v);
  Vec3Sub(v, x1);
  Vec3Norm(v);

  for (axisI = 0; axisI < 3; axisI++) {
    c = Vec3Scalar(v, dataP->vA[axisI]);
    if (c < 0.0)
      c = - c;
    dataP->val[3 * (dataP->bondI * dataP->molNo + dataP->molI) + axisI] =
      acosf(c) * 180.0 / M_PI;
  }

  dataP->bondI++;
}

static void
writeBondVal(DhBondP bondP, DhAtomP atom1P, DhAtomP atom2P, void *clientData)
{
  BondData *dataP = clientData;
  float *val;
  DSTR str;
  DhResP resP;
  char buf[20];
  float v, avg, der, min, max;
  int axisI, i;

  val = dataP->val + 3 * dataP->bondI * dataP->molNo;
  dataP->bondI++;

  str = DStrNew();
  resP = DhBondGetRes(bondP);

  (void) sprintf(buf, dataP->formRes, DhResGetNumber(resP));
  DStrAppStr(str, buf);
  (void) sprintf(buf, " %-5s", DStrToStr(DhResGetName(resP)));
  DStrAppStr(str, buf);
  (void) sprintf(buf, " %-5s", DStrToStr(DhAtomGetName(atom1P)));
  DStrAppStr(str, buf);
  (void) sprintf(buf, "-%-5s", DStrToStr(DhAtomGetName(atom2P)));
  DStrAppStr(str, buf);

  for (axisI = 0; axisI < 3; axisI++) {
    avg = 0.0;
    min = MAXFLOAT;
    max = - MAXFLOAT;

    for (i = 0; i < dataP->molNo; i++) {
      v = val[3 * i + axisI];
      avg += v;
      if (v < min)
	min = v;
      if (v > max)
	max = v;
    }

    avg /= dataP->molNo;

    if (dataP->molNo > 1) {
      der = 0.0;
      for (i = 0; i < dataP->molNo; i++) {
	v = val[3 * i + axisI];
	der += (v - avg) * (v - avg);
      }
      der = sqrtf(der / (dataP->molNo - 1));
    } else {
      der = 0.0;
    }

    if (axisI > 0)
      DStrAppStr(str, " ");
    (void) sprintf(buf, " %4.1f", avg);
    DStrAppStr(str, buf);
    (void) sprintf(buf, " +/- %4.1f", der);
    DStrAppStr(str, buf);
    (void) sprintf(buf, " (%4.1f", min);
    DStrAppStr(str, buf);
    (void) sprintf(buf, "..%4.1f)", max);
    DStrAppStr(str, buf);
  }

  DStrAppChar(str, '\n');
  PuWriteStr(TextW, DStrToStr(str));
  DStrFree(str);
}

static char *
getForm(int maxVal)
{
  if (maxVal >= 100000)
    return "%6d";

  if (maxVal >= 10000)
    return "%5d";

  if (maxVal >= 1000)
    return "%4d";

  if (maxVal >= 100)
    return "%3d";

  if (maxVal >= 10)
    return "%2d";

  return "%1d";
}

#define ARG_NUM 1
#define OPT_NUM 1

ErrCode
ExCalcAxis(char *cmd)
{
  ArgDescr arg[ARG_NUM];
  EnumEntryDescr enumEntry[OPT_NUM];
  ErrCode errCode;
  int molNo, molI;
  DhMolP *molPA;
  int maxAtomNo, atomNo, bondNo, maxResI;
  char *form, buf[20];
  PropRefP refP;
  AtomData atomData;
  BondData bondData;
  Vec3 cent;
  Mat4 mapM;
  float lenA[3], avg1, avg2;
  int i, k;

  arg[0].type = AT_MULT_ENUM;

  ArgInit(arg, ARG_NUM);

  enumEntry[0].str = "list bonds";
  enumEntry[0].onOff = ListBonds;

  arg[0].prompt = "Options";
  arg[0].u.enumD.entryP = enumEntry;
  arg[0].u.enumD.n = OPT_NUM;

  errCode = ArgGet(arg, ARG_NUM);
  if (errCode != EC_OK) {
    ArgCleanup(arg, ARG_NUM);
    return errCode;
  }

  ListBonds = enumEntry[0].onOff;

  DhActivateGroups(TRUE);

  molNo = SelMolGet(NULL, 0);
  if (molNo == 0) {
    DhActivateGroups(FALSE);
    CipSetError("at least one molecule must be selected");
    return EC_ERROR;
  }

  molPA = malloc(molNo * sizeof(*molPA));
  (void) SelMolGet(molPA, molNo);

  refP = PropGetRef(PROP_SELECTED, FALSE);

  maxAtomNo = 0;
  for (molI = 0; molI < molNo; molI++) {
    atomNo = 0;
    DhMolApplyAtom(refP, molPA[molI], countAtoms, &atomNo);
    if (atomNo > maxAtomNo)
      maxAtomNo = atomNo;
  }

  if (maxAtomNo == 0) {
    DhActivateGroups(FALSE);
    CipSetError("at least one atom must be selected");
    free(molPA);
    return EC_ERROR;
  }

  if (ListBonds) {
    bondData.bondNo = 0;
    DhMolApplyBond(refP, molPA[0], countBonds, &bondData.bondNo);
    for (molI = 1; molI < molNo; molI++) {
      bondNo = 0;
      DhMolApplyBond(refP, molPA[molI], countBonds, &bondNo);
      if (bondNo != bondData.bondNo) {
	DhActivateGroups(FALSE);
	CipSetError("same bonds must be selected in all structures");
	free(molPA);
	return EC_ERROR;
      }
    }
  }

  BreakActivate(TRUE);

  TextW = PuCreateTextWindow(cmd);

  PuWriteStr(TextW, "Relative lengths of principal axes:\n");

  PuWriteStr(TextW, "\n");
  form = getForm(molNo);
  for (molI = 0; molI < molNo; molI++) {
    (void) sprintf(buf, form, molI + 1);
    PuWriteStr(TextW, buf);
    PuWriteStr(TextW, " ");
    PuWriteStr(TextW, DStrToStr(DhMolGetName(molPA[molI])));
    PuWriteStr(TextW, "\n");
  }
  PuWriteStr(TextW, "\n");

  atomData.coordA = malloc(maxAtomNo * sizeof(Vec3));

  if (ListBonds) {
    bondData.molNo = molNo;

    maxResI = -1;
    DhMolApplyRes(refP, molPA[0], maxRes, &maxResI);
    bondData.formRes = getForm(maxResI);

    bondData.val = malloc(3 * molNo * bondData.bondNo * sizeof(*bondData.val));
  }

  avg1 = 0.0;
  avg2 = 0.0;

  for (molI = 0; molI < molNo; molI++) {
    if (BreakCheck(1))
      break;

    (void) sprintf(buf, form, molI + 1);
    PuWriteStr(TextW, buf);

    atomData.atomI = 0;
    DhMolApplyAtom(refP, molPA[molI], fillAtom, &atomData);

    if (atomData.atomI < 3) {
      PuWriteStr(TextW, "\n");
      continue;
    }

    Vec3Zero(cent);
    for (i = 0; i < atomData.atomI; i++)
      Vec3Add(cent, atomData.coordA[i]);
    Vec3Scale(cent, 1.0 / atomData.atomI);
    for (i = 0; i < atomData.atomI; i++)
      Vec3Sub(atomData.coordA[i], cent);

    PrincipAxis(atomData.coordA, atomData.atomI, FALSE, mapM, lenA);

    (void) sprintf(buf, " %5.3f", 1.0);
    PuWriteStr(TextW, buf);
    (void) sprintf(buf, " %5.3f", lenA[1] / lenA[0]);
    PuWriteStr(TextW, buf);
    (void) sprintf(buf, " %5.3f", lenA[2] / lenA[0]);
    PuWriteStr(TextW, buf);
    PuWriteStr(TextW, "\n");

    avg1 += lenA[1] / lenA[0];
    avg2 += lenA[2] / lenA[0];

    if (ListBonds) {
      bondData.molI = molI;
      bondData.bondI = 0;

      for (i = 0; i < 3; i++) {
	for (k = 0; k < 3; k++)
	  bondData.vA[i][k] = mapM[k][i];
	Vec3Norm(bondData.vA[i]);
      }

      DhMolApplyBond(refP, molPA[molI], getBondVal, &bondData);
    }
  }

  BreakActivate(FALSE);

  free(atomData.coordA);

  PuWriteStr(TextW, "\n");
  PuWriteStr(TextW, "average:");
  (void) sprintf(buf, " %5.3f", 1.0);
  PuWriteStr(TextW, buf);
  (void) sprintf(buf, " %5.3f", avg1 / molNo);
  PuWriteStr(TextW, buf);
  (void) sprintf(buf, " %5.3f", avg2 / molNo);
  PuWriteStr(TextW, buf);
  PuWriteStr(TextW, "\n");

  if (ListBonds) {
    PuWriteStr(TextW,
	"\nAngles of bonds to principal axes (longest..shortest):\n");

    bondData.bondI = 0;
    DhMolApplyBond(refP, molPA[0], writeBondVal, &bondData);

    free(bondData.val);
  }

  free(molPA);

  DhActivateGroups(FALSE);

  return EC_OK;
}
